# authentic2 - versatile identity manager
# Copyright (C) 2010-2022 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from authentic2.apps.authenticators.models import BaseAuthenticator
from authentic2.utils.misc import redirect_to_login

from . import views


class SAMLAuthenticator(BaseAuthenticator):
    metadata_url = models.URLField(_('Metadata URL'), max_length=300, blank=True)
    metadata_cache_time = models.PositiveSmallIntegerField(_('Metadata cache time'), default=3600)
    metadata_http_timeout = models.PositiveSmallIntegerField(_('Metadata HTTP timeout'), default=10)

    metadata_path = models.CharField(
        _('Metadata file path'),
        max_length=300,
        help_text=_('Absolute path to the IdP metadata file.'),
        blank=True,
    )
    metadata = models.TextField(_('Metadata (XML)'), blank=True)

    provision = models.BooleanField(
        _('Provision'),
        default=True,
        help_text=_('Create user if their username does not already exists.'),
    )
    verify_ssl_certificate = models.BooleanField(
        _('Verify SSL certificate'),
        default=True,
        help_text=_('Verify SSL certificate when doing HTTP requests, used when resolving artifacts.'),
    )
    transient_federation_attribute = models.CharField(
        _('Transient federation attribute'),
        max_length=64,
        help_text=_(
            'Name of an attribute to use in replacement of the NameID content when the NameID format is transient.'
        ),
        blank=True,
    )
    realm = models.CharField(
        _('Realm'),
        max_length=32,
        help_text=_('The default realm to associate to user, used when setting username.'),
        default='saml',
    )
    username_template = models.CharField(
        _('Username template'),
        max_length=128,
        help_text=_(
            'The template to build and/or retrieve a user from its username based '
            'on received attributes, the syntax is the one from the str.format() '
            'method of Python. Available variables are realm, idp (current settings '
            'for the idp issuing the assertion), attributes. The default value is '
            '{attributes[name_id_content]}@{realm}. Another example could be {atttributes[uid][0]} '
            'to set the passed username as the username of the newly created user.'
        ),
        default='{attributes[name_id_content]}@{realm}',
    )
    name_id_policy_format = models.CharField(
        _('NameID policy format'), max_length=64, help_text=_('The NameID format to request.'), blank=True
    )
    name_id_policy_allow_create = models.BooleanField(_('NameID policy allow create'), default=True)
    force_authn = models.BooleanField(
        _('Force authn'), default=False, help_text=_('Force authentication on each authentication request.')
    )
    add_authnrequest_next_url_extension = models.BooleanField(
        _('Add authnrequest next url extension'), default=False
    )
    group_attribute = models.CharField(
        _('Group attribute'),
        max_length=32,
        help_text=_('Name of the SAML attribute to map to Django group names (for example "role").'),
        blank=True,
    )
    create_group = models.BooleanField(
        _('Create group'), default=True, help_text=_('Create group or only assign existing groups.')
    )
    error_url = models.URLField(
        _('Error URL'),
        help_text=_(
            'URL for the continue link when authentication fails. If not set, the RelayState is '
            'used. If there is no RelayState, application default login redirect URL is used.'
        ),
        blank=True,
    )
    error_redirect_after_timeout = models.PositiveSmallIntegerField(
        _('Error redirect after timeout'),
        default=120,
        help_text=_(
            'Timeout in seconds before automatically redirecting the user to the '
            'continue URL when authentication has failed.'
        ),
    )
    authn_classref = models.CharField(
        _('Authn classref'),
        max_length=512,
        help_text=_(
            'Authorized authentication class references, separated by commas. '
            'Empty value means everything is authorized. Authentication class reference '
            'must be obtained from the identity provider but should come from the '
            'SAML 2.0 specification.'
        ),
        blank=True,
    )
    login_hints = models.CharField(
        _('Login hints'), max_length=512, help_text=_('Comma separated list.'), blank=True
    )
    lookup_by_attributes = JSONField(
        _('Lookup by attributes'),
        default=list,
        help_text=_(
            'Allow looking for user with some SAML attributes if the received NameID is '
            'still unknown. It must be a list of dictionnaries with two mandatory keys '
            '`user_field` and `saml_attribute`. The optionnal key `ignore-case` should be a '
            'boolean indicating if the match is case-insensitive (default is to respect the '
            'case).'
        ),
        blank=True,
    )
    a2_attribute_mapping = JSONField(
        _('Attribute mapping'),
        default=list,
        help_text=_(
            'Copy incoming SAML attributes to user attributes, for example '
            'for example [{"attribute": "email", "saml_attribute": "mail", "mandatory": false}].'
        ),
        blank=True,
    )
    attribute_mapping = JSONField(
        _('Attribute mapping (deprecated)'),
        default=dict,
        help_text=_(
            'Maps templates based on SAML attributes to field of the user model, '
            'for example {"email": "attributes[mail][0]"}.'
        ),
        blank=True,
    )
    superuser_mapping = JSONField(
        _('Superuser mapping'),
        default=dict,
        editable=False,
        help_text=_(
            'Gives superuser flags to user if a SAML attribute contains a given value, '
            'for example {"roles": "Admin"}.'
        ),
        blank=True,
    )

    type = 'saml'
    how = ['saml']
    description_fields = [
        'show_condition',
        'metadata_url',
        'metadata_path',
        'metadata',
        'provision',
        'username_template',
    ]

    class Meta:
        verbose_name = _('SAML')

    @property
    def settings(self):
        settings = {k.upper(): v for k, v in self.__dict__.items()}

        for setting in ('AUTHN_CLASSREF', 'LOGIN_HINTS'):
            settings[setting] = [x.strip() for x in settings[setting].split(',') if x.strip()]

        for setting in ('METADATA', 'METADATA_PATH', 'METADATA_URL'):
            if not settings[setting]:
                del settings[setting]

        return settings

    @property
    def manager_form_class(self):
        from .forms import SAMLAuthenticatorForm

        return SAMLAuthenticatorForm

    def clean(self):
        if not (self.metadata or self.metadata_path or self.metadata_url):
            raise ValidationError(_('One of the metadata fields must be filled.'))

    def autorun(self, request, block_id):
        from .adapters import AuthenticAdapter

        settings = self.settings
        AuthenticAdapter().load_idp(settings, self.order)
        return redirect_to_login(
            request, login_url='mellon_login', params={'entityID': settings['ENTITY_ID']}
        )

    def login(self, request, *args, **kwargs):
        return views.login(request, self, *args, **kwargs)

    def profile(self, request, *args, **kwargs):
        return views.profile(request, *args, **kwargs)
